import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import win32frame.Scrollbar;
import win32frame.TextBox;
import win32frame.UpDownCtrl;
import win32frame.ToolTip;
import win32frame.StaticControl;
import std.stdio;
import std.string;
import std.utf;
import std.conv;

enum CMD{
  UPDN  =0x301,
  TXTBOX=0x302
} ;

class TextBoxSpinEx : TextBox {
  this(string name) {
    super(name) ;
  }

  override LRESULT subproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    if( msg==WM_CHAR ){
      if( wp==VK_TAB || wp==VK_RETURN ){
        SetFocus(null) ;
        return(0) ;
      }
    }
    return(-1) ;
  }
}

class Spin : WindowObject {
  HGLRC     hRC;

  UpDownCtrl udcntl ;
  bool       update_text ;

  TextBoxSpinEx textbox ;
  string    textfmt ;
  double    cur, max, min, step ;
  int       scrW = 12 ;

  StaticControl label ;
  string    labelStr = "" ;
  int       labelW   = 0 ;

  this(string name, string ClassName) {
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;

    cur= 0.0 ;
    min= 1.0 ;
    max=-1.0 ;
    step=1.0 ;
    textfmt="%.0f" ;
    update_text=false ;
  }

  int setInfo(double cur, double min, double max, double step){
    this.cur = cur ;
    this.min = min ;
    this.max = max ;
    this.step= step ;
    return(0) ;
  }

  int setValue(double cur){
    if( min<=max ){
      cur = (cur<min) ? min : (cur>max) ? max : cur ;
    }
    if( textbox.getHWND() != null ){
      textbox.setWindowText(format(textfmt,cur)) ;
    }
    this.cur=cur ;
    return(0) ;
  }

  double getValue(){
    return(cur) ;
  }

  int setTextFormat(string fmt){
    textfmt=fmt ;
    return(0) ;
  }

  int setTextFont(HFONT hfont){
    if( getHWND() != null ){
      return(textbox.setFont(hfont)) ;
    }
    return(-1) ;
  }

  int setLabel(string label, int width){
    this.labelStr  = label ;
    if( width>=0 ){
      this.labelW = width ;
    }
    return(0) ;
  }

  int setLabelFont(HFONT hfont){
    if( getHWND() != null ){
      return(label.setFont(hfont)) ;
    }
    return(-1) ;
  }

  int setLabelLeftAlign(){
    if( getHWND() != null ){
      LONG style=GetWindowLong(label.getHWND(), GWL_STYLE) ;
      SetWindowLong(label.getHWND(), GWL_STYLE, (style & ~3)|SS_LEFT ) ;
      return(0) ;
    }
    return(-1) ;
  }

  int setLabelCenterAlign(){
    if( getHWND() != null ){
      LONG style=GetWindowLong(label.getHWND(), GWL_STYLE) ;
      SetWindowLong(label.getHWND(), GWL_STYLE, (style & ~3)|SS_CENTER ) ;
      return(0) ;
    }
    return(-1) ;
  }

  override int addToolTip(ToolTip TooltipOBJ){
    if( (TooltipOBJ !is null) && (udcntl !is null) ){
      tooltip = TooltipOBJ ;
      int ret = tooltip.create(udcntl.getHWND()) ;
      tooltip.setClientRect() ;
      return(ret) ;
    }
    return(-1) ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      break ;

    case WM_CREATE:
      udcntl = new UpDownCtrl("updowncontrol") ;
      udcntl.setPosition(winW-scrW,winY,scrW,winH) ;
      addObject(udcntl, cast(HMENU)(CMD.UPDN), false, false) ;
      udcntl.setRange(0,100) ;
      udcntl.setPos(50) ;

      int tabstop = getStyle() & WS_TABSTOP ;
      string defaultText = format(textfmt,cur) ;
      textbox= new TextBoxSpinEx(defaultText) ;
      textbox.addExStyle(WS_EX_CLIENTEDGE) ;
      textbox.addStyle(WS_BORDER | ES_RIGHT | ES_AUTOHSCROLL | tabstop) ;
      textbox.setPosition(winX+labelW,winY,winW-(scrW+labelW),winH);
      addObject(textbox, cast(HMENU)(CMD.TXTBOX), false, false) ;
      label = new StaticControl(labelStr) ;
      label.addStyle(SS_RIGHT|SS_CENTERIMAGE) ;
      label.setPosition(winX,winY,labelW,winH) ;
      addObject(label, null) ;
      break ;
    case WM_SIZE :
      if( (udcntl !is null) && (textbox !is null) && (label !is null) ){
        MoveWindow(  udcntl.getHWND(), winW-scrW, 0,               scrW,winH,true) ;
        MoveWindow( textbox.getHWND(), labelW,    0, winW-(scrW+labelW),winH,true) ;
        MoveWindow(   label.getHWND(),      0,    0,             labelW,winH,true) ;
      }
      break ;

    case WM_VSCROLL:
      bool   changed=false ;
      double d=0.0 ;
      if( HIWORD(wp)<50 ){
        d =  step * (50 - HIWORD(wp));
        changed=true ;
      }else if( HIWORD(wp)>50 ){
        d = -step * (HIWORD(wp) -50 );
        changed=true ;
      }
      udcntl.setPos(50) ;
      if( changed==true ){
        SetFocus(null) ;
        SetFocus(textbox.getHWND()) ;
        cur += d ;
        if( min<=max ){
          cur = (cur<min) ? min : (cur>max) ? max : cur ;
        }
        textbox.setWindowText(format(textfmt,cur)) ;
        sendmsgtoparent( WM_COMMAND, cast(WPARAM)(getHMENU()), 0) ;
      }
      break ;

    case WM_SETFOCUS :
      SetFocus(textbox.getHWND()) ;
      sendmsg(textbox.getHWND(), EM_SETSEL, 0,65535) ;
      break ;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case CMD.TXTBOX:
        if( HIWORD(wp)==EN_KILLFOCUS ){
          string val=textbox.getWindowText() ;
          if( isNumeric(val)==true ){
            cur=to!(double)(val) ;
            if( min<=max ){
              cur = (cur<min) ? min : (cur>max) ? max : cur ;
            }
          }
          textbox.setWindowText(format(textfmt,cur)) ;
          sendmsgtoparent( WM_COMMAND, cast(WPARAM)(getHMENU()), 0) ;
        }
        break ;
      default : break ;
      }
      break ;

    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

}
